use crate::radio::MAX_PACKET_LEN;
use anyhow::anyhow;
use colored::{Color, Colorize};
use ham_cats::{
    packet::Packet,
    whisker::{Route, RouteNode},
};

// RF to RF, also Internet to RF
pub fn should_digipeat(
    callsign: &str,
    ssid: u8,
    packet: &Packet<MAX_PACKET_LEN>,
    echo: bool,
) -> bool {
    let route = match packet.route() {
        Some(x) => x,
        None => {
            if echo {
                println!("Not digipeating - Packet has no route")
            };
            return false;
        }
    };

    if let Some(ident) = packet.identification() {
        if &ident.callsign == callsign && ident.ssid == ssid {
            // this node is the source of the packet
            if echo {
                println!("Not digipeating - Packet ident is us");
            }
            return false;
        }
    }

    let max_hops: usize = route.max_hops.into();
    let cur_hops = route
        .iter()
        .filter(|r| match r {
            RouteNode::Internet => false,
            RouteNode::Identity(_, _, is_future) => !is_future,
        })
        .count();

    if max_hops <= cur_hops {
        // already hit our max hops
        if echo {
            println!("Not digipeating - Max hops hit");
        }
        return false;
    }

    let already_digipeated = route.iter().any(|r| match r {
        RouteNode::Internet => false,
        RouteNode::Identity(rc, rs, is_future) => rc == callsign && rs == ssid && !is_future,
    });

    if already_digipeated {
        // this node has already digipeated this packet
        if echo {
            println!("Not digipeating - Already digipeated by this node")
        };
        return false;
    }

    let next_node = route.iter().find_map(|r| match r {
        RouteNode::Identity(c, s, is_future) if is_future => Some((c, s)),
        _ => None,
    });

    match next_node {
        Some((rc, rs)) => {
            let out = rc == callsign && rs == ssid;
            if !out && echo {
                println!("Not digipeating - This packet's destiny is set and does not involve us");
            }
            out
        }
        None => true,
    }
}

pub fn print_packet(pkt: &Packet<MAX_PACKET_LEN>, is_rf: bool) {
    let color = if is_rf {
        println!("-----  RF -----");

        Color::Green
    } else {
        println!("--- FELINET ---");

        Color::Magenta
    };

    if let Some(ident) = pkt.identification() {
        println!(
            "{} \t{}-{} [ICON: {}]",
            "IDENT:".color(color),
            ident.callsign,
            ident.ssid,
            ident.icon
        );
    }
    if let Some(route) = pkt.route() {
        let r_str = route
            .iter()
            .map(|rn| match rn {
                ham_cats::whisker::RouteNode::Internet => "[NET]".to_string(),
                ham_cats::whisker::RouteNode::Identity(rc, rs, is_future) => {
                    let s = if is_future { "*" } else { "" };
                    format!("{rc}-{rs}{s}")
                }
            })
            .collect::<Vec<String>>()
            .join(" -> ");
        println!(
            "{} \t(MAX {}) {r_str}",
            "ROUTE:".color(color),
            route.max_hops
        );
    }

    let mut buf = [0; MAX_PACKET_LEN];
    if let Ok(comment) = pkt.comment(&mut buf) {
        println!("{} \t'{comment}'", "CMNT:".color(color));
    }
    if let Some(gps) = pkt.gps() {
        println!(
            "{} \t({:07.2}, {:07.2}) +/- {:03} m, v = {:.2} m/s [N {:.2} deg]",
            "GPS:".color(color),
            gps.latitude(),
            gps.longitude(),
            gps.max_error,
            gps.speed,
            gps.heading()
        );
    }
    println!();
}

pub fn attempt_decode(packet: Vec<u8>) -> anyhow::Result<Packet<MAX_PACKET_LEN>> {
    let packet: Packet<MAX_PACKET_LEN> =
        Packet::fully_decode(&packet).map_err(|e| anyhow!("Could not decode packet: {e:?}"))?;

    print_packet(&packet, true);

    Ok(packet)
}

// i-gating is kind of weird.
// if a packet has a pre-determined route (is_future=true) and our node
// doesn't match it, we don't digipeat it.
// But FELINET should be an "all-seeing eye". Any packet that's heard
// should end up on the network.
// The solution is that, if the packet has non-matching is_future=true nodes,
// We rip them off, and stick our node + an internet route on
#[must_use]
pub fn append_internet_to_packet_route(
    callsign: &str,
    ssid: u8,
    packet: &mut Packet<MAX_PACKET_LEN>,
) -> Option<()> {
    let mut route = packet.route().unwrap_or(Route::new(0));

    let first_future = route.iter().find_map(|rn| match rn {
        RouteNode::Identity(rc, rs, is_future) if is_future => Some((rc, rs)),
        _ => None,
    });

    if let Some((rc, rs)) = first_future {
        if rc == callsign && rs == ssid {
            // we can just do the normal thing
            append_to_route(callsign, ssid, &mut route)?;
        } else {
            // rip off all the futures and add ourselves
            let mut new_route = Route::new(route.max_hops);
            for r in route.iter() {
                match r {
                    RouteNode::Internet => new_route.push_internet()?,
                    RouteNode::Identity(rc, rs, is_future) => {
                        if is_future {
                            break;
                        }

                        new_route.push_callsign(rc, rs, false);
                    }
                }
            }

            route = new_route;
            route.push_callsign(callsign, ssid, false);
        }
    } else {
        append_to_route(callsign, ssid, &mut route)?;
    }

    route.push_internet()?;
    packet.clear_route();
    packet.add_route(route).ok()?;

    Some(())
}

#[must_use]
pub fn append_to_packet_route(
    callsign: &str,
    ssid: u8,
    packet: &mut Packet<MAX_PACKET_LEN>,
) -> Option<()> {
    let mut route = packet.route().unwrap_or(Route::new(0));
    append_to_route(callsign, ssid, &mut route)?;
    packet.clear_route();
    packet.add_route(route).ok()?;

    Some(())
}

#[must_use]
fn append_to_route(callsign: &str, ssid: u8, r: &mut Route) -> Option<()> {
    let replace_future = r.iter().any(|rn| match rn {
        RouteNode::Identity(rc, rs, is_future) => rc == callsign && rs == ssid && is_future,
        _ => false,
    });

    if replace_future {
        let mut new_route = Route::new(r.max_hops);
        let mut already_replaced = false;

        for rn in r.iter() {
            match rn {
                RouteNode::Identity(rc, rs, is_future)
                    if rc == callsign && rs == ssid && is_future && !already_replaced =>
                {
                    already_replaced = true;
                    new_route.push_callsign(callsign, ssid, false)?;
                }
                RouteNode::Identity(rc, rs, is_future) => {
                    new_route.push_callsign(rc, rs, is_future)?;
                }
                RouteNode::Internet => {
                    new_route.push_internet()?;
                }
            }
        }

        *r = new_route;
    } else {
        r.push_callsign(callsign, ssid, false)?;
    }

    Some(())
}
